from ocdsmerge.flatten import flatten


def test_flatten_1():  # from documentation
    data = {
        "c": "I am a",
        "b": ["A", "list"],
        "a": [
            {"id": 1, "cb": "I am ca"},
            {"id": 2, "ca": "I am cb"}
        ]
    }

    assert flatten(data, {}, {}) == {
        ('c',): 'I am a',
        ('b',): ['A', 'list'],
        ('a', '1', 'cb'): 'I am ca',
        ('a', '1', 'id'): 1,
        ('a', '2', 'ca'): 'I am cb',
        ('a', '2', 'id'): 2,
    }


def test_flatten_2():
    data = {
        "a": [
            {"id": "identifier"},
            {"key": "value"}
        ]
    }

    actual = flatten(data, {}, {})
    keys = list(actual.keys())
    values = list(actual.values())

    assert len(actual) == 2
    assert values == ['identifier', 'value']
    assert keys[0] == ('a', 'identifier', 'id')
    assert len(keys[1]) == 3
    assert keys[1][0] == 'a'
    assert len(keys[1][1]) == 36
    assert keys[1][2] == 'key'



import os
import warnings

import pytest

from ocdsmerge import APPEND, MERGE_BY_POSITION, Merger
from ocdsmerge.errors import DuplicateIdValueWarning
from tests import load

releases = load(os.path.join('schema', 'identifier-merge-duplicate-id.json'))


@pytest.mark.vcr()
def test_warn(empty_merger):
    fields = ['identifierMerge', 'array']
    string = "Multiple objects have the `id` value '1' in the `nested.{}` array"

    with pytest.warns(DuplicateIdValueWarning) as records:
        empty_merger.create_compiled_release(releases)

    assert len(records) == 2

    for i, record in enumerate(records):
        message = string.format(fields[i])

        assert record.message.path == ('nested', fields[i],)
        assert record.message.id == '1'
        assert record.message.message == message
        assert str(record.message) == message


@pytest.mark.vcr()
def test_raise(empty_merger):
    with pytest.raises(DuplicateIdValueWarning) as excinfo:
        with warnings.catch_warnings():
            warnings.filterwarnings('error', category=DuplicateIdValueWarning)
            empty_merger.create_compiled_release(releases)

    message = "Multiple objects have the `id` value '1' in the `nested.identifierMerge` array"

    assert excinfo.value.path == ('nested', 'identifierMerge',)
    assert excinfo.value.id == '1'
    assert excinfo.value.message == message
    assert str(excinfo.value) == message


@pytest.mark.vcr()
def test_ignore(empty_merger):
    with pytest.warns(None) as records:
        with warnings.catch_warnings():
            warnings.filterwarnings('ignore', category=DuplicateIdValueWarning)
            empty_merger.create_compiled_release(releases)

    assert not records, 'unexpected warning: {}'.format(records[0].message)


@pytest.mark.vcr()
def test_merge_by_position():
    merger = Merger(rule_overrides={('nested', 'array',): MERGE_BY_POSITION})

    with pytest.warns(DuplicateIdValueWarning):
        compiled_release = merger.create_compiled_release(releases + releases)

    assert compiled_release == load(os.path.join('schema', 'identifier-merge-duplicate-id-by-position.json'))


@pytest.mark.vcr()
def test_append():
    merger = Merger(rule_overrides={('nested', 'array',): APPEND})

    with pytest.warns(DuplicateIdValueWarning):
        compiled_release = merger.create_compiled_release(releases + releases)

    assert compiled_release == load(os.path.join('schema', 'identifier-merge-duplicate-id-append.json'))



class OCDSMergeError(Exception):
    """Base class for exceptions from within this package"""


class MissingDateKeyError(OCDSMergeError, KeyError):
    """Raised when a release is missing a 'date' key"""

    def __init__(self, key, message):
        self.key = key
        self.message = message

    def __str__(self):
        return str(self.message)


class NonObjectReleaseError(OCDSMergeError, TypeError):
    """Raised when a release is not an object"""


class NullDateValueError(OCDSMergeError, TypeError):
    """Raised when a release has a null 'date' value"""


class NonStringDateValueError(OCDSMergeError, TypeError):
    """Raised when a release has a non-string 'date' value"""


class OCDSMergeWarning(UserWarning):
    """Base class for warnings from within this package"""


class DuplicateIdValueWarning(OCDSMergeWarning):
    """Used when at least two objects in the same array have the same value for the 'id' field"""

    def __init__(self, path, id, message):
        self.path = path
        self.id = id
        self.message = message

    def __str__(self):
        return str(self.message)



import json
import os.path
import re
from copy import deepcopy
from glob import glob

import pytest

from ocdsmerge import CompiledRelease, Merger, VersionedRelease
from ocdsmerge.errors import MissingDateKeyError, NonObjectReleaseError, NonStringDateValueError, NullDateValueError
from tests import load, path, schema_url, tags

simple_schema = load('schema.json')

test_merge_argvalues = []
for minor_version, schema in (('1.1', None), ('1.1', schema_url), ('1.0', schema_url), ('schema', simple_schema)):
    if isinstance(schema, str):
        schema = schema.format(tags[minor_version])
    for suffix in ('compiled', 'versioned'):
        filenames = glob(path(os.path.join(minor_version, '*-{}.json'.format(suffix))))
        assert len(filenames), '{} fixtures not found'.format(suffix)
        test_merge_argvalues += [(filename, schema) for filename in filenames]


@pytest.mark.vcr()
@pytest.mark.parametrize('error, data', [
    (MissingDateKeyError, {}),
    (NullDateValueError, {'date': None}),
    (NonStringDateValueError, {'date': {}}),
    (NonObjectReleaseError, '{}'),
    (NonObjectReleaseError, b'{}'),
    (NonObjectReleaseError, []),
    (NonObjectReleaseError, tuple()),
    (NonObjectReleaseError, set()),
])
def test_errors(error, data, empty_merger):
    for infix in ('compiled', 'versioned'):
        with pytest.raises(error):
            getattr(empty_merger, 'create_{}_release'.format(infix))([{'date': '2010-01-01'}, data])

    if not isinstance(data, dict):
        with pytest.raises(error):
            empty_merger.create_compiled_release([data])
    else:
        release = deepcopy(data)

        expected = {
            'id': 'None-{}'.format(data.get('date')),
            'tag': ['compiled'],
        }

        if data.get('date') is not None:
            expected['date'] = data['date']

        assert empty_merger.create_compiled_release([release]) == expected

    if not isinstance(data, dict):
        with pytest.raises(error):
            empty_merger.create_versioned_release([data])
    else:
        release = deepcopy(data)
        release['initiationType'] = 'tender'

        expected = {
            'initiationType': [{
                'releaseID': None,
                'releaseDate': data.get('date'),
                'releaseTag': None,
                'value': 'tender',
            }],
        }

        assert empty_merger.create_versioned_release([release]) == expected


@pytest.mark.vcr()
@pytest.mark.parametrize('filename,schema', test_merge_argvalues)
def test_merge(filename, schema):
    merger = Merger(schema)

    if filename.endswith('-compiled.json'):
        infix = 'compiled'
    else:
        infix = 'versioned'

    with open(filename) as f:
        expected = json.load(f)
    with open(re.sub(r'-(?:compiled|versioned)', '', filename)) as f:
        releases = json.load(f)

    original = deepcopy(releases)
    actual = getattr(merger, 'create_{}_release'.format(infix))(releases)

    assert releases == original
    assert actual == expected, filename + '\n' + json.dumps(actual)


@pytest.mark.vcr()
@pytest.mark.parametrize('infix,cls', [('compiled', CompiledRelease), ('versioned', VersionedRelease)])
def test_extend(infix, cls, empty_merger):
    releases = load(os.path.join('1.1', 'lists.json'))

    merged_release = getattr(empty_merger, 'create_{}_release'.format(infix))(releases[:1])

    merger = cls(merged_release, merge_rules=empty_merger.merge_rules)
    merger.extend(releases[1:])

    expected = load(os.path.join('1.1', 'lists-{}.json'.format(infix)))

    assert merger.asdict() == expected


@pytest.mark.vcr()
@pytest.mark.parametrize('infix,cls', [('compiled', CompiledRelease), ('versioned', VersionedRelease)])
def test_append(infix, cls, empty_merger):
    releases = load(os.path.join('1.1', 'lists.json'))

    merged_release = getattr(empty_merger, 'create_{}_release'.format(infix))(releases[:1])

    merger = cls(merged_release, merge_rules=empty_merger.merge_rules)
    merger.append(releases[1])

    expected = load(os.path.join('1.1', 'lists-{}.json'.format(infix)))

    assert merger.asdict() == expected


@pytest.mark.parametrize('i,j', [(0, 0), (0, 1), (1, 0), (1, 1)])
def test_merge_when_array_is_mixed(i, j, simple_merger):
    data = [{
        "ocid": "ocds-213czf-A",
        "id": "1",
        "date": "2000-01-01T00:00:00Z",
        "mixedArray": [
            {"id": 1},
            "foo"
        ]
    }, {
        "ocid": "ocds-213czf-A",
        "id": "2",
        "date": "2000-01-02T00:00:00Z",
        "mixedArray": [
            {"id": 2},
            "bar"
        ]
    }]

    output = {
        'tag': ['compiled'],
        'id': 'ocds-213czf-A-2000-01-02T00:00:00Z',
        'date': '2000-01-02T00:00:00Z',
        'ocid': 'ocds-213czf-A',
        'mixedArray': [
            {'id': 2},
            'bar',
        ],
    }

    assert simple_merger.create_compiled_release(data) == output

    actual = deepcopy(data)
    expected = deepcopy(output)
    del actual[i]['mixedArray'][j]
    if i == 1:
        del expected['mixedArray'][j]

    assert simple_merger.create_compiled_release(actual) == expected, \
        'removed item index {} from release index {}'.format(j, i)


@pytest.mark.parametrize('i,j', [(0, 0), (0, 1), (1, 0), (1, 1)])
def test_merge_when_array_is_mixed_without_schema(i, j, empty_merger):
    data = [{
        'ocid': 'ocds-213czf-A',
        "id": "1",
        "date": "2000-01-01T00:00:00Z",
        "mixedArray": [
            {"id": 1},
            "foo"
        ]
    }, {
        'ocid': 'ocds-213czf-A',
        "id": "2",
        "date": "2000-01-02T00:00:00Z",
        "mixedArray": [
            {"id": 2},
            "bar"
        ]
    }]

    output = {
        'tag': ['compiled'],
        'id': 'ocds-213czf-A-2000-01-02T00:00:00Z',
        'date': '2000-01-02T00:00:00Z',
        'ocid': 'ocds-213czf-A',
        'mixedArray': [
            {'id': 2},
            'bar',
        ],
    }

    assert empty_merger.create_compiled_release(data) == output

    actual = deepcopy(data)
    expected = deepcopy(output)
    del actual[i]['mixedArray'][j]
    if i == 1:
        del expected['mixedArray'][j]

    if j == 0:
        assert empty_merger.create_compiled_release(actual) == expected, \
            'removed item index {} from release index {}'.format(j, i)
    else:
        with pytest.raises(AssertionError):
            assert empty_merger.create_compiled_release(actual) == expected, \
                'removed item index {} from release index {}'.format(j, i)



from ocdsmerge.util import get_release_schema_url, get_tags


def test_get_release_schema_url():
    assert get_release_schema_url('1__1__3') >= 'https://standard.open-contracting.org/schema/1__1__3/release-schema.json'  # noqa


def test_get_tags():
    assert get_tags()[:11] == [
        '0__3__2',
        '0__3__3',
        '1__0__0',
        '1__0__1',
        '1__0__2',
        '1__0__3',
        '1__1__0',
        '1__1__1',
        '1__1__2',
        '1__1__3',
        '1__1__4',
    ]



from ocdsmerge.rules import get_merge_rules
from tests import schema_url, tags


def test_get_merge_rules_1_1():
    assert get_merge_rules(schema_url.format(tags['1.1'])) == {
        ('awards', 'items', 'additionalClassifications'): {'wholeListMerge'},
        ('contracts', 'items', 'additionalClassifications'): {'wholeListMerge'},
        ('contracts', 'relatedProcesses', 'relationship'): {'wholeListMerge'},
        ('date',): {'omitWhenMerged'},
        ('id',): {'omitWhenMerged'},
        ('parties', 'additionalIdentifiers'): {'wholeListMerge'},
        ('parties', 'roles'): {'wholeListMerge'},
        ('relatedProcesses', 'relationship'): {'wholeListMerge'},
        ('tag',): {'omitWhenMerged'},
        ('tender', 'additionalProcurementCategories'): {'wholeListMerge'},
        ('tender', 'items', 'additionalClassifications'): {'wholeListMerge'},
        ('tender', 'submissionMethod'): {'wholeListMerge'},

        # Deprecated
        ('awards', 'amendment', 'changes'): {'wholeListMerge'},
        ('awards', 'amendments', 'changes'): {'wholeListMerge'},
        ('awards', 'suppliers', 'additionalIdentifiers'): {'wholeListMerge'},
        ('buyer', 'additionalIdentifiers'): {'wholeListMerge'},
        ('contracts', 'amendment', 'changes'): {'wholeListMerge'},
        ('contracts', 'amendments', 'changes'): {'wholeListMerge'},
        ('contracts', 'implementation', 'transactions', 'payee', 'additionalIdentifiers'): {'wholeListMerge'},
        ('contracts', 'implementation', 'transactions', 'payer', 'additionalIdentifiers'): {'wholeListMerge'},
        ('tender', 'amendment', 'changes'): {'wholeListMerge'},
        ('tender', 'amendments', 'changes'): {'wholeListMerge'},
        ('tender', 'procuringEntity', 'additionalIdentifiers'): {'wholeListMerge'},
        ('tender', 'tenderers', 'additionalIdentifiers'): {'wholeListMerge'},
    }


def test_get_merge_rules_1_0():
    assert get_merge_rules(schema_url.format(tags['1.0'])) == {
        ('awards', 'amendment', 'changes'): {'wholeListMerge'},
        ('awards', 'items', 'additionalClassifications'): {'wholeListMerge'},
        ('awards', 'suppliers'): {'wholeListMerge'},
        ('buyer', 'additionalIdentifiers'): {'wholeListMerge'},
        ('contracts', 'amendment', 'changes'): {'wholeListMerge'},
        ('contracts', 'items', 'additionalClassifications'): {'wholeListMerge'},
        ('date',): {'omitWhenMerged'},
        ('id',): {'omitWhenMerged'},
        ('ocid',): {'omitWhenMerged'},
        ('tag',): {'omitWhenMerged'},
        ('tender', 'amendment', 'changes'): {'wholeListMerge'},
        ('tender', 'items', 'additionalClassifications'): {'wholeListMerge'},
        ('tender', 'procuringEntity', 'additionalIdentifiers'): {'wholeListMerge'},
        ('tender', 'submissionMethod'): {'wholeListMerge'},
        ('tender', 'tenderers'): {'wholeListMerge'},
    }



"""
It's possible to regenerate fixtures with commands like:

cat tests/fixtures/1.1/contextual.json | jq -crM .[] | ocdskit package-releases | ocdskit --pretty compile \
    > tests/fixtures/1.1/contextual-compiled.json
cat tests/fixtures/1.1/contextual.json | jq -crM .[] | ocdskit package-releases | ocdskit --pretty compile \
    --versioned > tests/fixtures/1.1/contextual-versioned.json
cat tests/fixtures/1.0/suppliers.json  | jq -crM .[] | ocdskit package-releases | ocdskit --pretty compile \
    --versioned --schema https://standard.open-contracting.org/schema/1__0__3/release-schema.json \
    > tests/fixtures/1.0/suppliers-versioned.json
"""

import json
import os
import warnings
from glob import glob

import pytest
from jsonschema import FormatChecker
from jsonschema.validators import Draft4Validator as validator

from tests import load, path, tags

release_schema_path = 'release-schema-{}.json'
versioned_release_schema_path = 'versioned-release-validation-schema-{}.json'

test_valid_argvalues = []
for minor_version, patch_tag in tags.items():
    filenames = glob(path(os.path.join(minor_version, '*.json')))
    assert len(filenames), 'ocds fixtures not found'
    for versioned, schema_path in ((False, release_schema_path), (True, versioned_release_schema_path)):
        schema = load(schema_path.format(patch_tag))
        for filename in filenames:
            if not versioned ^ filename.endswith('-versioned.json'):
                test_valid_argvalues.append((filename, schema))


def custom_warning_formatter(message, category, filename, lineno, line=None):
    return str(message).replace(os.getcwd() + os.sep, '')


warnings.formatwarning = custom_warning_formatter


@pytest.mark.parametrize('filename,schema', test_valid_argvalues)
def test_valid(filename, schema):
    errors = 0

    with open(filename) as f:
        data = json.load(f)
    if filename.endswith('-versioned.json') or filename.endswith('-compiled.json'):
        data = [data]

    for datum in data:
        for error in validator(schema, format_checker=FormatChecker()).iter_errors(datum):
            errors += 1
            warnings.warn(json.dumps(error.instance, indent=2, separators=(',', ': ')))
            warnings.warn('{} ({})\n'.format(error.message, '/'.join(error.absolute_schema_path)))

    assert errors == 0, '{} is invalid. See warnings below.'.format(filename)
